﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Runtime.InteropServices;
using System.IO;
using System.Threading;
using Microsoft.Win32.SafeHandles;
using Microsoft.Win32;

namespace ConsoleApp6
{
    class Program
    {
        // 调试模式
        public const Boolean DEBUGMODE = false;

        [DllImport("kernel32", SetLastError = true)]
        static extern IntPtr CreateFile(
                                string FileName,
                                uint DesiredAccess,
                                uint ShareMode,
                                IntPtr SecurityAttributes,
                                uint CreationDisposition,
                                int FlagsAndAttributes,
                                IntPtr hTemplate
                                );

        static class DESIREDACCESS
        {
            public const uint GENERIC_READ = 0x80000000;
            public const uint GENERIC_WRITE = 0x40000000;
            public const uint GENERIC_EXECUTE = 0x20000000;
            public const uint GENERIC_ALL = 0x10000000;
        }
        /// <summary>
        /// Sharing mode of the file or object
        ///</summary>
        static class SHAREMODE
        {
            public const uint FILE_SHARE_READ = 0x00000001;
            public const uint FILE_SHARE_WRITE = 0x00000002;
            public const uint FILE_SHARE_DELETE = 0x00000004;
        }
        /// <summary>
        /// Action to take on files that exist, and which action to take when files do not exist.
        /// </summary>
        static class CREATIONDISPOSITION
        {
            public const uint CREATE_NEW = 1;
            public const uint CREATE_ALWAYS = 2;
            public const uint OPEN_EXISTING = 3;
            public const uint OPEN_ALWAYS = 4;
            public const uint TRUNCATE_EXISTING = 5;
        }
        /// <summary>
        /// File attributes and flags for the file.
        /// </summary>
        static class FLAGSANDATTRIBUTES
        {
            public const uint FILE_FLAG_WRITE_THROUGH = 0x80000000;
            public const uint FILE_FLAG_OVERLAPPED = 0x40000000;
            public const uint FILE_FLAG_NO_BUFFERING = 0x20000000;
            public const uint FILE_FLAG_RANDOM_ACCESS = 0x10000000;
            public const uint FILE_FLAG_SEQUENTIAL_SCAN = 0x08000000;
            public const uint FILE_FLAG_DELETE_ON_CLOSE = 0x04000000;
            public const uint FILE_FLAG_BACKUP_SEMANTICS = 0x02000000;
            public const uint FILE_FLAG_POSIX_SEMANTICS = 0x01000000;
            public const uint FILE_FLAG_OPEN_REPARSE_POINT = 0x00200000;
            public const uint FILE_FLAG_OPEN_NO_RECALL = 0x00100000;
            public const uint FILE_FLAG_FIRST_PIPE_INSTANCE = 0x00080000;
        }

        // 用于 32MB SPI NOR
        public const uint MAXSPIADDRESS = 32 * 1024 * 1024;
        // 用于  8MB PSRAM
        public const uint MAXPSRAMADDRESS = 8 * 1024 * 1024;
        // 用于 256MB Disk
        public const uint MAXDISKADDRESS = 256 * 1024 * 1024;
        // 一次写入数据量
        public const uint ONECLUSTERSIZE = 4 * 1024;
        // 一个扇区的大小
        public const int ONESECTORSIZE = 512;

        //结构体1字节对齐
        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        unsafe struct TranHeader
        {
            public UInt32 Flag;
            public UInt32 Monot;
            public UInt32 Size;
            public UInt32 Counter;
            public UInt16 FilenameLength;
            public fixed byte Filename[494];
        };
        // 以 obj 的长度进行比较
        static bool EqualsBytes(ref byte[] Dest, ref byte[] Source)
        {
            for (int i = 0; i < Source.Length; i++)
            {
                if (Dest[i] != Source[i])
                    return false;
            }
            return true;
        }


        /// <summary>
        /// 由结构体转换为byte数组
        /// </summary>
        public static byte[] StructureToByte<T>(T structure)
        {
            int size = Marshal.SizeOf(typeof(T));
            byte[] buffer = new byte[size];
            IntPtr bufferIntPtr = Marshal.AllocHGlobal(size);
            try
            {
                Marshal.StructureToPtr(structure, bufferIntPtr, false);
                Marshal.Copy(bufferIntPtr, buffer, 0, size);
            }
            finally
            {
                Marshal.FreeHGlobal(bufferIntPtr);
            }
            return buffer;
        }

        static void Main(string[] args)
        {
            // 当前硬盘信息在
            // Computer\HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\disk\Enum

            RegistryKey  myreg = Registry.LocalMachine.OpenSubKey(@"SYSTEM\CurrentControlSet\Services\disk\Enum");
            // 如果当前系统只有1个硬盘，那么直接退出
            if (myreg.GetValue("Count").ToString() == "1") {
                myreg.Close();
                Console.WriteLine("No transfer disk, pressing anykey to exit");
                Console.ReadKey();
                Environment.Exit(0);
            }

            string[] subkeyNames = myreg.GetValueNames();
            int DiskNo=0;
            foreach (string keyName in subkeyNames)
            {
                if (myreg.GetValue(keyName).ToString().StartsWith(@"USBSTOR\Disk&Ven_DUMT")) {
                    if (DEBUGMODE)
                    {
                        Console.WriteLine(keyName);
                        Console.WriteLine(myreg.GetValue(keyName).ToString());
                    }
                    DiskNo = int.Parse(keyName);
                }
            }
            myreg.Close();
            if (DiskNo == 0) {
                Console.WriteLine("No transfer disk, pressing anykey to exit");
                Console.ReadKey();
                Environment.Exit(0);
            }

            byte[] Sector = new byte[ONECLUSTERSIZE];

            IntPtr ptr = CreateFile(@"\\.\PHYSICALDRIVE"+ DiskNo.ToString(),
                            DESIREDACCESS.GENERIC_READ | DESIREDACCESS.GENERIC_WRITE,
                            SHAREMODE.FILE_SHARE_READ | SHAREMODE.FILE_SHARE_WRITE,
                            IntPtr.Zero,
                            CREATIONDISPOSITION.OPEN_EXISTING,
                            0,
                            IntPtr.Zero);
            SafeFileHandle PD1handle = new SafeFileHandle(ptr, true);
            FileStream disk1 = new FileStream(PD1handle, FileAccess.ReadWrite);

            // 检查第5个扇区，判断是否为设备
            disk1.Seek(0x800, SeekOrigin.Begin);
            disk1.Read(Sector, 0, ONESECTORSIZE);
            Console.WriteLine("Checking PHYSICALDRIVE" + DiskNo.ToString()+" Signature:");
            string arr = "DATTRANSFTERDISKWWW.LAB-Z.COM";
            byte[] bytearr = System.Text.Encoding.Default.GetBytes(arr);

            // 如果签名不符合，说明不是我们需要的设备
            if (Program.EqualsBytes(ref Sector, ref bytearr) != true)
            {
                //不是UDT设备，输出错误然后退出
                Console.WriteLine("PHYSICALDRIVE"+DiskNo.ToString() + " is NOT a DATATRANFER! ");
                disk1.Close();
                Console.ReadLine();
                Environment.Exit(0);
            }
            else
            {
                Console.WriteLine("PHYSICALDRIVE" + DiskNo.ToString() + " signature tests pass! ");
            }

            // 检查是否收到命令行参数
            if (args.Count() != 0)
            {
                // 判断命令行参数给出的文件名是否存在
                if (!File.Exists(args[0]))
                {
                    Console.WriteLine("Can't find {0} file!", args[0]);
                    Console.ReadLine();
                    Environment.Exit(0);
                }
            }

            if (DEBUGMODE)
            {
                Console.WriteLine("First Read:");
                for (int i = 0; i < ONESECTORSIZE; i++)
                {
                    Console.Write("{0:x} ", Sector[i]);
                }
                Console.WriteLine(" ");
            }

            // 检查 Busy 位
            disk1.Seek(0x1800, SeekOrigin.Begin);
            disk1.Read(Sector, 0, 4096);
            if (DEBUGMODE)
            {
                Console.WriteLine("Busy check:");
                for (int i = 0; i < 32; i++)
                {
                    Console.Write("{0:x} ", Sector[i]);
                }
                Console.WriteLine(" ");
            }

            //检查 Busy位
            if ((Sector[0] & 1) == 0)
            {
                Console.WriteLine("Device busy, please try again later!");
                Console.ReadLine();
                Environment.Exit(0);
            }

            if (DEBUGMODE)
            {
                Console.WriteLine("Device is free!");
            }

            var tr = new TranHeader();

            tr.Flag = (uint)(Sector[0] + (Sector[1] << 8) + (Sector[2] << 16) + (Sector[3] << 24));
            tr.Monot = (uint)(Sector[4] + (Sector[5] << 8) + (Sector[6] << 16) + (Sector[7] << 24));

            //取得要写入的文件信息
            FileStream fi = new FileStream(args[0], FileMode.Open, FileAccess.Read);

            //取文件长度
            tr.Size = (UInt32)fi.Length;
            //文件数量
            tr.Counter = 0;

            // 赋值文件名
            byte[] fn = System.Text.Encoding.Default.GetBytes(args[0]);
            for (int i = 0; i < fn.Length; i++)
            {
                unsafe
                {
                    tr.Filename[i] = fn[i];
                }
            }

            //文件名长度
            tr.FilenameLength = (UInt16)fn.Length;
            //Ziv
            if (DEBUGMODE)
            {
                Console.WriteLine("Flag: {0:x}\r\nMono: {1:x}\r\nSize: {2}(D)\r\nCounter: {3}(D)",
                                tr.Flag,
                                tr.Monot,
                                tr.Size,
                                tr.Counter
                                );
            }
            else {
                Console.WriteLine("File Size: {0}(D)",tr.Size);
            }
            
            Console.OutputEncoding = System.Text.Encoding.ASCII;
            Console.Write("Filename:");
            for (int i = 0; i < tr.FilenameLength; i++)
            {
                unsafe
                {
                    Console.Write((char)tr.Filename[i]);
                }
            }
            Console.WriteLine(" ");

            // 自增标志加一
            tr.Monot++;

            // 设置 Busy位
            tr.Flag = tr.Flag & 0xFFFFFFFE;
            // 回写，标明现在正在写入操作
            Sector = StructureToByte(tr);
            byte[] Cluster = new byte[ONECLUSTERSIZE];
            Array.Copy(Sector, Cluster, 512);
            disk1.Seek(0x1800, SeekOrigin.Begin);
            disk1.Write(Cluster, 0, 512);

            byte[] tmp = new byte[Cluster.Length - Sector.Length];

            // 已经处理的字节数
            int BytesLeft = (int)fi.Length;
            int Counter = 0;

            while (BytesLeft > 0)
            {
                try
                {
                    // 取当前剩下的和4K两个中最小的
                    fi.Read(Cluster, 0, Math.Min(BytesLeft, (int)ONECLUSTERSIZE));
                    disk1.Seek(MAXDISKADDRESS - MAXPSRAMADDRESS + (Counter * ONECLUSTERSIZE), SeekOrigin.Begin);
                    disk1.Write(Cluster, 0, (int)ONECLUSTERSIZE);
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Send file error");
                    throw ex;
                }
                BytesLeft = BytesLeft - Math.Min(BytesLeft, (int)ONECLUSTERSIZE);
                Counter++;
            }


            //清除 Busy 位
            tr.Flag = tr.Flag | 0x1;
            Sector = StructureToByte(tr);
            disk1.Seek(0x1800, SeekOrigin.Begin);
            disk1.Write(Sector, 0, 512);

            disk1.Close();
            Console.WriteLine("Done, exit");
            //Console.ReadKey();
        }
    }
}
